/*
 * Config.h has the following entries
 * // Replace with your network credentials
 *const char* ssid = "SSID";
 *const char* password = "PASSWORD";

 * // Initialize Telegram BOT
 *#define BOTtoken "BOTTOKEN"  // your Bot Token (Get from Botfather)

 * // Use @myidbot to find out the chat ID of an individual or a group
 * // Also note that you need to click "start" on a bot before it can
 * // message you
 *#define CHAT_ID "CHATID"
 */
 
#include <FastLED.h>
#include <Wire.h>
#include <Adafruit_BMP280.h>

#ifdef ESP32
#include <WiFi.h>
#else
#include <ESP8266WiFi.h>
#endif
#include <WiFiClientSecure.h>
#include <UniversalTelegramBot.h>
#include <ArduinoJson.h>

#include "DHTesp.h"

//Wifi & Telgram credentials
#include "config.h"

// Sensors & Control

#define pin_DHT11 4 //
#define pin_alarmsensor 2 //RCWL-0516
#define pin_siren 19
#define pin_led_veranda 26
#define pin_led_bedroom 25
#define pin_doorbell 27

DHTesp dht;

#define DATA_PIN    18
#define LED_TYPE    WS2811
#define COLOR_ORDER GRB
#define NUM_LEDS   12
CRGB leds[NUM_LEDS];

#define BRIGHTNESS          100
#define FRAMES_PER_SECOND  120

bool alarmArmed = false;
bool alarmTriggered = false;

bool partyledson = false;
bool livingroomledson = false;

int buttonState;
int lastButtonState = LOW;
unsigned long lastDebounceTime = 0;
unsigned long debounceDelay = 50;

bool buttonPushed = false;

WiFiClientSecure client;
UniversalTelegramBot bot(BOTtoken, client);

int botRequestDelay = 50;
unsigned long lastTimeBotRan;

char textoption;

String chat_id;

TaskHandle_t loopingTasks;
TaskHandle_t nonloopingTasks;
SemaphoreHandle_t  xMutex;

void GetTemperature() { //implemented
  String message = "Temperature: ";
  TempAndHumidity newValues = dht.getTempAndHumidity();
  message.concat(String(newValues.temperature)); message.concat(".");
  bot.sendMessage(chat_id, message, "");
}

void ToggleAlarm() {
  alarmArmed = !alarmArmed;

  if (alarmArmed) {
    bot.sendMessage(chat_id, "Alarm armed.", "");
  }
  else {
    alarmArmed = false;
    alarmTriggered = false;
    bot.sendMessage(chat_id, "Alarm disarmed.", "");
  }

}

void TogglePartylights() {
  partyledson = !partyledson;
  if (partyledson) {
    livingroomledson = false;
    bot.sendMessage(chat_id, "Party lights switched on.", "");
  }
  else bot.sendMessage(chat_id, "Party lights switched off.", "");
}

void ToggleVerandaLights() {
  bool LightStatus = digitalRead(pin_led_veranda);
  LightStatus = !LightStatus;
  digitalWrite(pin_led_veranda, LightStatus);
  LightStatus = digitalRead(pin_led_veranda);
  if (LightStatus) {
    bot.sendMessage(chat_id, "Veranda light switched on.", "");
  }
  else bot.sendMessage(chat_id, "Veranda light switched off.", "");
}

void ToggleLivingRoomLights() {
  partyledson = false;
  bool LightStatus = livingroomledson;
  livingroomledson = !livingroomledson;
  if (livingroomledson) {
    for ( int i = 0; i < NUM_LEDS; i++) {
      leds[i] = CRGB::White;
    }
    FastLED.show();
    bot.sendMessage(chat_id, "Livingroom light switched on.", "");
  }
  else {
    FastLED.clear();
    FastLED.show();
    bot.sendMessage(chat_id, "Livingroom light switched off.", "");
  }

}

void ToggleBedroomLights() {
  bool LightStatus = digitalRead(pin_led_bedroom);
  LightStatus = !LightStatus;
  digitalWrite(pin_led_bedroom, LightStatus);
  LightStatus = digitalRead(pin_led_bedroom);
  if (LightStatus) {
    bot.sendMessage(chat_id, "Bedroom lights switched on.", "");
  }
  else bot.sendMessage(chat_id, "Bedroom lights switched off.", "");

}

void soundAlarm() {
  digitalWrite(pin_siren, HIGH);
  delay(500);
  digitalWrite(pin_siren, LOW);
  delay(500);
}

void GetSystemStatus() {
  String message = "";

  if (alarmArmed) {
    message += "Alarm is armed";
  }
  else message += "Alarm is disarmed.\n";
  if (digitalRead(pin_siren)) {
    message += "and sounding.\n";
  }
  else if (alarmArmed) message += ".\n";

  if (digitalRead(pin_led_bedroom)) {
    message += "Bedroom lights are on.\n";
  }
  else message += "Bedroom lights are off.\n";

  if (partyledson) {
    message += "Party lights are on.\n";
  }
  else message += "Partylights are off off.\n";

  if (livingroomledson) {
    message += "Livingroom lights are on.\n";
  }
  else message += "Livingroom lights are off.\n";

  message += "Temperature: ";
  TempAndHumidity newValues = dht.getTempAndHumidity();
  message.concat(String(newValues.temperature)); message.concat("ºC.\n");

  if (digitalRead(pin_led_veranda)) {
    message += "Veranda lights are on.\n";
  }
  else message += "Veranda lights are off.\n";

  bot.sendMessage(chat_id, message, "");
}

// Handle what happens when you receive new messages
void handleNewMessages(int numNewMessages) {
  Serial.println("handleNewMessages");
  Serial.println(String(numNewMessages));

  for (int i = 0; i < numNewMessages; i++) {
    // Chat id of the requester
    chat_id = String(bot.messages[i].chat_id);
    if (chat_id != CHAT_ID) {
      bot.sendMessage(chat_id, "Unauthorized user", "");
      Serial.println("Unauthorized user");
      continue;
    }

    // Print the received message
    String text = bot.messages[i].text;
    Serial.println(text);

    String from_name = bot.messages[i].from_name;
    if (text == "/temperature") {
      textoption = 't';
    }
    else if (text == "/status") {
      textoption = 's';
    }
    else if (text == "/alarm") {
      textoption = 'a';
    }
    else if (text == "/party") {
      textoption = 'p';
    }
    else if (text == "/veranda_lights") {
      textoption = 'v';
    }
    else if (text == "/livingroom_lights") {
      textoption = 'l';
    }
    else if (text == "/bedroom_lights") {
      textoption = 'b';
    }
    else textoption = 'x';

    switch (textoption) {
      case 't' : {
          GetTemperature();
          break;
        }
      case 's' : {
          GetSystemStatus();
          break;
        }
      case 'a' : {
          ToggleAlarm();
          break;
        }
      case 'p' : {
          TogglePartylights();
          break;
        }
      case 'v' : {
          ToggleVerandaLights();
          break;
        }
      case 'l' : {
          ToggleLivingRoomLights();
          break;
        }
      case 'b' : {
          ToggleBedroomLights();
          break;
        }
      default : {
          bot.sendMessage(chat_id, "Invalid command. \nUse one of:\n\/alarm\n/bedroom_lights\n/livingroom_lights\n/party\n/temperature\n/veranda_lights", "");
          break;
        }
    }
  }
}
void setup() {
  delay(3000);

  Serial.begin(115200);

  dht.setup(pin_DHT11, DHTesp::DHT11);
#ifdef ESP8266
  client.setInsecure();
#endif

  FastLED.addLeds<LED_TYPE, DATA_PIN, COLOR_ORDER>(leds, NUM_LEDS).setCorrection(TypicalLEDStrip);
  FastLED.setBrightness(BRIGHTNESS);
  FastLED.clear();
  FastLED.show();
  Serial.println("Started");

  pinMode(pin_siren, OUTPUT);
  pinMode(pin_led_veranda, OUTPUT);
  pinMode(pin_led_bedroom, OUTPUT);
  pinMode(pin_alarmsensor, INPUT);
  pinMode(pin_doorbell, INPUT_PULLUP);

  // Connect to Wi-Fi
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  client.setCACert(TELEGRAM_CERTIFICATE_ROOT);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }
  // Print ESP32 Local IP Address
  Serial.println(WiFi.localIP());

  Serial.println("Bot started?");
  bot.sendMessage(CHAT_ID, "System restarted", "");

  xMutex = xSemaphoreCreateMutex();

  //create a task that will be executed in the Task1code() function, with priority 1 and executed on core 0
  xTaskCreatePinnedToCore(
    nonloopingCode,   /* Task function. */
    "nonloopingTasks",     /* name of task. */
    16384,       /* Stack size of task */
    NULL,        /* parameter of the task */
    6,           /* priority of the task */
    &nonloopingTasks,      /* Task handle to keep track of created task */
    0);          /* pin task to core 0 */
  delay(500);

  xTaskCreatePinnedToCore(
    loopingCode,   /* Task function. */
    "loopingTasks",     /* name of task. */
    16384,       /* Stack size of task */
    NULL,        /* parameter of the task */
    6,           /* priority of the task */
    &loopingTasks,      /* Task handle to keep track of created task */
    1);          /* pin task to core 1 */
  delay(500);

}

void checkTelegramBot() {
  if (millis() > lastTimeBotRan + botRequestDelay) {
    int numNewMessages = bot.getUpdates(bot.last_message_received + 1);

    while (numNewMessages) {
      Serial.println("Got response");
      handleNewMessages(numNewMessages);
      numNewMessages = bot.getUpdates(bot.last_message_received + 1);
    }
    lastTimeBotRan = millis();
  }
}

void checkDoorbell() {
  int reading = digitalRead(pin_doorbell);
  if (reading != lastButtonState) {
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) > debounceDelay) {
    if (reading != buttonState) {
      buttonState = reading;
      if (buttonState == LOW) {
        soundAlarm(); soundAlarm();
        buttonPushed = true;
      }
    }
  }
  lastButtonState = reading;
}

void nonloopingCode(void * pvParamters) {
  Serial.print("nonloopingCode running on core ");
  Serial.println(xPortGetCoreID());
  soundAlarm();
  for (;;) {

    if ((!partyledson) && (!livingroomledson)) {
      FastLED.clear();
      FastLED.show();
    }
    if (alarmArmed) {
      if (digitalRead(pin_alarmsensor)) {
        int motionSensor = digitalRead(pin_alarmsensor); //Read Pin as input
        if ((motionSensor > 0) && (!alarmTriggered))
        {
          Serial.println("Motion Detected");
          bot.sendMessage(chat_id, "Alarm triggered", "");
          partyledson = false;
          alarmTriggered = true;
        }
      }
    }
    xSemaphoreTake(xMutex, portMAX_DELAY);
    if (buttonPushed) {
      bot.sendMessage(chat_id, "Someone is at the door.", "");
      buttonPushed = false;
    }
    checkTelegramBot();
    xSemaphoreGive(xMutex); // release mutex
  }
}
void loopingCode(void * pvParamters) {
  Serial.print("Juggle running on core ");
  Serial.println(xPortGetCoreID());
  Serial.println("Up and running");
  delay(1000);
  soundAlarm();
  for (;;) {
    checkDoorbell();
    if (alarmTriggered)soundAlarm();
    if (partyledson) {
      fadeToBlackBy( leds, NUM_LEDS, 20);
      byte dothue = 0;
      int i = 0;
      while ((i < 8) && partyledson) {
        leds[beatsin16( i + 7, 0, NUM_LEDS - 1 )] |= CHSV(dothue, 200, 255);
        dothue += 32;
        i += 1;
      }
      // send the 'leds' array out to the actual LED strip
      if (!partyledson) FastLED.clear();
      FastLED.show();
      // insert a delay to keep the framerate modest
    }
  }
}

void loop() {
  vTaskDelete(NULL);
}
